@@ -1,9 +1,9 @@
 """
 TODO: write doc
 """
-__title__ = 'FLI camera modules'
+__title__ = "FLI camera modules"
 
 from .flicamera import FliCamera
 
 
-__all__ = ['FliCamera']
+__all__ = ["FliCamera"]
